import x10.util.Timer;
import x10.io.Console;

class laplacian {

    public static def println(s:String) = Console.ERR.println(s);
    public static def print(s:String) = Console.ERR.print(s);
    public static def now():double = Timer.nanoTime() * 1e-9;
    public static def format(x:double, numDecimals:int) {
	return String.format("%1."+numDecimals+"f", [x as Box[Double]]);
    }

    global u : Array[double](2);
    global v : Array[double](2);
    val n : Int;
    global lambda : Double;

    public def this(m:Int)
    {
	n = m;
        lambda = +0.025;

	var R : Region(2) = [0..m+1,0..m+1];
	var D : Dist(2) = Dist.makeBlock(R, 0);
	u = Array.make[double](D, (p:Point) => { p(0) == 0 ? 1.0 : 0.0 });
	v = Array.make[double](D, (p:Point) => { p(0) == 0 ? 1.0 : 0.0 });    
    }

    public def output()
    {
	var i:int;
	var j:int;

	println("");
	for (i=0; i<n+2; i++) {
	    for (j=0; j<n+2; j++) {
		val ii = i;
		val jj = j;
		at (u.dist(i,j))
		    print(" " + format(u(ii, jj),6));
	    }
	    println("");
	}
    }

    public def iterate()
    {
        val m = n;

        finish 
	for (p:Place in Place.places)
           async (p) {
	       
	       val i0 = (u.dist | here).region.min(0);
	       val i1 = (u.dist | here).region.max(0);

	       if (i0 > 0) {
		   val w = at (p.prev()) 
		       ValRail.make[double](m+2,(j:Int)=>u(i0-1,j));
		   for (var j:int = 1; j<m+1; j++)
		       v(i0,j) = (1-4*lambda) * u(i0,j) + lambda* 
			   (u(i0+1,j) + w(j) + u(i0,j-1) + u(i0,j+1));
	       }
	       else {
		   for (var j:int = 1; j<m+1; j++)
		       v(1,j) = (1-4*lambda) * u(1,j) + lambda* 
			   (u(2,j) + u(0,j) + u(1,j-1) + u(1,j+1));
	       }

	       if (i1 < m+1) {
		   val w = at (p.next()) 
		       ValRail.make[double](m+2,(j:Int)=>u(i1+1,j));
		   for (var j:int = 1; j<m+1; j++)
		       v(i1, j) = (1-4*lambda) * u(i1,j) + lambda* 
			   (w(j) + u(i1-1,j) + u(i1,j-1) + u(i1,j+1));
	       }
	       else {
		   for (var j:int = 1; j<m+1; j++)
		       v(m, j) = (1-4*lambda) * u(m,j) + lambda* 
			   (u(m+1,j) + u(m-1,j) + u(m,j-1) + u(m,j+1));
	       }
	       
	       for (var i:int=i0+1; i<i1; i++)
		   for (var j:int=1; j<m+1; j++)
		       v(i,j) = (1-4*lambda) * u(i,j) + lambda* 
			   (u(i+1,j) + u(i-1,j) + u(i,j-1) + u(i,j+1));
	   }


	var i:int;
	var j:int;
	var dmax : Double = -1.0;
	var d : Double = 0.0;
	for (i=1; i<n+1; i++) {
	    for (j=1; j<n+1; j++) {
		val ii = i;
		val jj = j;
		d = Math.abs(at (u.dist(ii,jj)) ( u(ii,jj) - v(ii,jj) ));
		if (d > dmax) dmax = d;
		at (u.dist(i,j))
		    u(ii, jj) = v(ii,jj);
	    }
	}
	return dmax;
    }

    public static def main(args:Rail[String]!) {

	var n:Int = (args.length > 0) ? int.parseInt(args(0)) : 10;
	var err:Double = 1.0;
	var it:Int = 0;

	val l = new laplacian(n);
	l.output();

	var t:double = now();

	while ((err > 1e-3) && (it < 10)) {
	    it = it + 1;
	    err = l.iterate();
	    println("it = " + it + " err = " + err);
	}
	println("time elapsed : " + (now() - t));

	l.output();
   }

}
